#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# #    Copyright 2023 Thomas Schmelzer
# #
# #    Licensed under the Apache License, Version 2.0 (the "License");
# #    you may not use this file except in compliance with the License.
# #    You may obtain a copy of the License at
# #
# #        http://www.apache.org/licenses/LICENSE-2.0
# #
# #    Unless required by applicable law or agreed to in writing, software
# #    distributed under the License is distributed on an "AS IS" BASIS,
# #    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# #    See the License for the specific language governing permissions and
# #    limitations under the License.
"""
Popular year vs month performance table.

Supports compounded and cumulative (i.e. fixed AUM) returns logic.
"""
from __future__ import annotations

import calendar
from enum import Enum

import numpy as np
import pandas as pd


def _compound_returns(returns):
    return (1.0 + returns).prod() - 1.0


def _cumulative_returns(returns):
    return returns.sum()


class Aggregate(Enum):
    COMPOUND = _compound_returns
    CUMULATIVE = _cumulative_returns


def monthlytable(returns: pd.Series, f: Aggregate) -> pd.DataFrame:
    """
    Get a table of monthly returns.

    Args:
        returns: Series of individual returns.
        f: Aggregate function to use.

    Returns:
        DataFrame with monthly returns, their STDev and YTD.
    """
    # Works better in the first month
    # Compute all the intramonth-returns
    # instead of reapplying some monthly resampling of the NAV
    r = pd.Series(returns)

    return_monthly = r.groupby([r.index.year, r.index.month]).apply(f)

    frame = return_monthly.unstack(level=1).rename(
        columns=lambda x: calendar.month_abbr[x]
    )

    ytd = frame.apply(f, axis=1)
    frame["STDev"] = np.sqrt(12) * frame.std(axis=1)
    # make sure that you don't include the column for the STDev in your computation
    frame["YTD"] = ytd
    frame.index.name = "Year"
    frame.columns.name = None
    # most recent years on top
    return frame.iloc[::-1]



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from __future__ import annotations

from dataclasses import dataclass
from typing import Generator

import numpy as np
import pandas as pd

from cvx.simulator.portfolio import Portfolio
from cvx.simulator.state import State
from cvx.simulator.utils.interpolation import valid
from cvx.simulator.utils.rescale import returns2prices


@dataclass
class Builder:
    """
    The Builder is an auxiliary class used to build portfolios.
    It overloads the __iter__ method to allow the class to iterate over
    the timestamps for which the portfolio data is available.

    In each iteration we can update the portfolio by setting either
    the weights, the position or the cash position.

    After the iteration has been completed we build a Portfolio object
    by calling the build method.
    """

    prices: pd.DataFrame
    initial_aum: float = 1e6

    _state: State = None
    _units: pd.DataFrame = None
    _aum: pd.Series = None

    def __post_init__(self) -> None:
        """
        The __post_init__ method is a special method of initialized instances
        of the _Builder class and is called after initialization.
        It sets the initial amount of cash in the portfolio to be equal to the input initial_cash parameter.

        The method takes no input parameter. It initializes the cash attribute in the internal
        _State object with the initial amount of cash in the portfolio, self.initial_cash.

        Note that this method is often used in Python classes for additional initialization routines
        that can only be performed after the object is fully initialized. __post_init__
        is called automatically after the object initialization.
        """

        # assert isinstance(self.prices, pd.DataFrame)
        assert self.prices.index.is_monotonic_increasing
        assert self.prices.index.is_unique

        self._state = State()

        self._units = pd.DataFrame(
            index=self.prices.index,
            columns=self.prices.columns,
            data=np.NaN,
            dtype=float,
        )

        self._aum = pd.Series(index=self.prices.index, dtype=float)

        self._state.aum = self.initial_aum

    @property
    def valid(self):
        """
        Analyse the validity of the data
        Do it column by column of the prices
        """
        return self.prices.apply(valid)

    @property
    def intervals(self):
        """
        Find for each column the first and the last valid index
        """
        return self.prices.apply(
            lambda ts: pd.Series(
                {"first": ts.first_valid_index(), "last": ts.last_valid_index()}
            )
        ).transpose()

    @property
    def index(self) -> pd.DatetimeIndex:
        """A property that returns the index of the portfolio,
        which are the timestamps for which the portfolio data is available.

        Returns: pd.Index: A pandas index representing the
        time period for which the portfolio data is available.
        """
        return pd.DatetimeIndex(self.prices.index)

    @property
    def current_prices(self) -> np.array:
        """
        Get the current prices from the state
        """
        return self._state.prices[self._state.assets].values

    def __iter__(self) -> Generator[tuple[pd.DatetimeIndex, State], None, None]:
        """
        The __iter__ method allows the object to be iterated over in a for loop,
        yielding time and the current state of the portfolio.
        The method yields a list of dates seen so far and returns a tuple
        containing the list of dates and the current portfolio state.

        Yield:

        time: a pandas DatetimeIndex object containing the dates seen so far.
        state: the current state of the portfolio,

        taking into account the stock prices at each interval.
        """
        for t in self.index:
            # update the current prices for the portfolio
            self._state.prices = self.prices.loc[t]

            # update the current time for the state
            self._state.time = t

            # yield the vector of times seen so far and the current state
            yield self.index[self.index <= t], self._state

    @property
    def position(self) -> pd.Series:
        """
        The position property returns the current position of the portfolio.
        It returns a pandas Series object containing the current position of the portfolio.

        Returns: pd.Series: a pandas Series object containing the current position of the portfolio.
        """
        return self._units.loc[self._state.time]

    @position.setter
    def position(self, position: pd.Series) -> None:
        """
        The position property returns the current position of the portfolio.
        It returns a pandas Series object containing the current position of the portfolio.

        Returns: pd.Series: a pandas Series object containing the current position of the portfolio.
        """
        self._units.loc[self._state.time, self._state.assets] = position
        self._state.position = position

    @property
    def cashposition(self):
        """
        The cashposition property returns the current cash position of the portfolio.
        """
        return self.position * self.current_prices

    @property
    def units(self):
        """
        The units property returns the frame of holdings of the portfolio.
        Useful mainly for testing
        """
        return self._units

    @cashposition.setter
    def cashposition(self, cashposition: pd.Series) -> None:
        """
        The cashposition property sets the current cash position of the portfolio.
        """
        self.position = cashposition / self.current_prices

    def build(self):
        """A function that creates a new instance of the EquityPortfolio
        class based on the internal state of the Portfolio builder object.

        Returns: EquityPortfolio: A new instance of the EquityPortfolio class
        with the attributes (prices, units, initial_cash, trading_cost_model) as specified in the Portfolio builder.

        Notes: The function simply creates a new instance of the EquityPortfolio
        class with the attributes (prices, units, initial_cash, trading_cost_model) equal
        to the corresponding attributes in the Portfolio builder object.
        The resulting EquityPortfolio object will have the same state as the Portfolio builder from which it was built.
        """

        return Portfolio(prices=self.prices, units=self.units, aum=self.aum)

    @property
    def weights(self) -> np.array:
        """
        Get the current weights from the state
        """
        return self._state.weights[self._state.assets].values

    @weights.setter
    def weights(self, weights: np.array) -> None:
        """
        The weights property sets the current weights of the portfolio.
        We convert the weights to positions using the current prices and the NAV
        """
        self.position = self._state.nav * weights / self.current_prices

    @property
    def aum(self):
        """
        The aum property returns the current AUM of the portfolio.
        """
        return self._aum

    @aum.setter
    def aum(self, aum):
        """
        The aum property sets the current AUM of the portfolio.
        """
        self._aum[self._state.time] = aum
        self._state.aum = aum

    @classmethod
    def from_returns(cls, returns):
        """Build Futures Portfolio from returns"""
        # compute artificial prices (but scaled such their returns are correct)

        prices = returns2prices(returns)
        return cls(prices=prices)



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from cvx.simulator.utils.metric import sharpe
from cvx.simulator.utils.rescale import returns2prices


@dataclass(frozen=True)
class Portfolio:
    prices: pd.DataFrame
    units: pd.DataFrame
    aum: float | pd.Series

    def __post_init__(self) -> None:
        """A class method that performs input validation after object initialization.
        Notes: The post_init method is called after an instance of the Portfolio
        class has been initialized, and performs a series of input validation
        checks to ensure that the prices and units dataframes are in the
        expected format with no duplicates or missing data,
        and that the units dataframe represents valid equity positions
        for the assets held in the portfolio.
        Specifically, the method checks that both the prices and units dataframes
        have a monotonic increasing and unique index,
        and that the index and columns of the units dataframe are subsets
        of the index and columns of the prices dataframe, respectively.
        If any of these checks fail, an assertion error will be raised."""

        assert self.prices.index.is_monotonic_increasing
        assert self.prices.index.is_unique
        assert self.units.index.is_monotonic_increasing
        assert self.units.index.is_unique

        assert set(self.units.index).issubset(set(self.prices.index))
        assert set(self.units.columns).issubset(set(self.prices.columns))

    @property
    def index(self) -> pd.DatetimeIndex:
        """A property that returns the index of the EquityPortfolio instance,
        which is the time period for which the portfolio data is available.

        Returns: pd.Index: A pandas index representing the time period for which the
        portfolio data is available.

        Notes: The function extracts the index of the prices dataframe,
        which represents the time periods for which data is available for the portfolio.
        The resulting index will be a pandas index object with the same length
        as the number of rows in the prices dataframe."""
        return pd.DatetimeIndex(self.prices.index)

    @property
    def assets(self) -> pd.Index:
        """A property that returns a list of the assets held by the EquityPortfolio object.

        Returns: list: A list of the assets held by the EquityPortfolio object.

        Notes: The function extracts the column names of the prices dataframe,
        which correspond to the assets held by the EquityPortfolio object.
        The resulting list will contain the names of all assets held by the portfolio, without any duplicates.
        """
        return self.prices.columns

    @property
    def nav(self) -> pd.Series:
        """Return a pandas series representing the NAV"""
        if isinstance(self.aum, pd.Series):
            series = self.aum
        else:
            profit = (self.cashposition.shift(1) * self.returns.fillna(0.0)).sum(axis=1)
            series = profit.cumsum() + self.aum

        series.name = "NAV"
        return series

    @property
    def profit(self) -> pd.Series:
        """A property that returns a pandas series representing the
        profit gained or lost in the portfolio based on changes in asset prices.

        Returns: pd.Series: A pandas series representing the profit
        gained or lost in the portfolio based on changes in asset prices.
        """
        series = (self.cashposition.shift(1) * self.returns.fillna(0.0)).sum(axis=1)
        series.name = "Profit"
        return series

    @property
    def highwater(self) -> pd.Series:
        """A function that returns a pandas series representing
        the high-water mark of the portfolio, which is the highest point
        the portfolio value has reached over time.

        Returns: pd.Series: A pandas series representing the
        high-water mark of the portfolio.

        Notes: The function performs a rolling computation based on
        the cumulative maximum of the portfolio's value over time,
        starting from the beginning of the time period being considered.
        Min_periods argument is set to 1 to include the minimum period of one day.
        The resulting series will show the highest value the portfolio has reached at each point in time.
        """
        series = self.nav.expanding(min_periods=1).max()
        series.name = "Highwater"
        return series

    @property
    def drawdown(self) -> pd.Series:
        """A property that returns a pandas series representing the
        drawdown of the portfolio, which measures the decline
        in the portfolio's value from its (previously) highest
        point to its current point.

        Returns: pd.Series: A pandas series representing the
        drawdown of the portfolio.

        Notes: The function calculates the ratio of the portfolio's current value
        vs. its current high-water-mark and then subtracting the result from 1.
        A positive drawdown means the portfolio is currently worth
        less than its high-water mark. A drawdown of 0.1 implies that the nav is currently 0.9 times the high-water mark
        """
        series = 1.0 - self.nav / self.highwater
        series.name = "Drawdown"
        return series

    @property
    def cashposition(self):
        """
        Return a pandas dataframe representing the cash position
        """
        return self.prices * self.units

    @property
    def returns(self):
        """
        The returns property exposes the returns of the individual assets
        """
        return self.prices.pct_change()

    @property
    def trades_units(self) -> pd.DataFrame:
        """A property that returns a pandas dataframe representing the trades made in the portfolio in terms of units.

        Returns: pd.DataFrame: A pandas dataframe representing the trades made in the portfolio in terms of units.

        Notes: The function calculates the trades made by the portfolio by taking
        the difference between the current and previous values of the units dataframe.
        The resulting values will represent the number of shares of each asset
        bought or sold by the portfolio at each point in time.
        The resulting dataframe will have the same dimensions
        as the units dataframe, with NaN values filled with zeros."""
        t = self.units.fillna(0.0).diff()
        t.loc[self.index[0]] = self.units.loc[self.index[0]]
        return t.fillna(0.0)

    @property
    def trades_currency(self) -> pd.DataFrame:
        """A property that returns a pandas dataframe representing
        the trades made in the portfolio in terms of currency.

        Returns: pd.DataFrame: A pandas dataframe representing the trades made in the portfolio in terms of currency.

        Notes: The function calculates the trades made in currency by multiplying
        the number of shares of each asset bought or sold (as represented in the trades_units dataframe)
        with the current prices of each asset (as represented in the prices dataframe).
        The resulting dataframe will have the same dimensions as the units and prices dataframes.
        """
        return self.trades_units * self.prices

    @property
    def turnover_relative(self) -> pd.DataFrame:
        """A property that returns a pandas dataframe representing the turnover
        relative to the NAV. Can be positive (if bought) or negative (if sold)."""
        return self.trades_currency.div(self.nav, axis=0)

    @property
    def turnover(self) -> pd.DataFrame:
        return self.trades_currency.abs()

    def __getitem__(self, time: datetime) -> pd.Series:
        """The `__getitem__` method retrieves the stock data for a specific time in the dataframe.
        It returns the stock data for that time.

        The method takes one input parameter:
        - `time`: the time index for which to retrieve the stock data

        Returns:
        - stock data for the input time

        Note that the input time must be in the index of the dataframe,
        otherwise a KeyError will be raised."""
        return self.units.loc[time]

    @property
    def equity(self) -> pd.DataFrame:
        """A property that returns a pandas dataframe
        representing the equity positions of the portfolio,
        which is the value of each asset held by the portfolio.
        Returns: pd.DataFrame: A pandas dataframe representing
        the equity positions of the portfolio.

        Notes: The function calculates the equity of the portfolio
        by multiplying the current prices of each asset
        by the number of shares held by the portfolio.
        The equity dataframe will have the same dimensions
        as the prices and units dataframes."""

        return self.cashposition

    @property
    def weights(self) -> pd.DataFrame:
        """A property that returns a pandas dataframe representing
        the weights of various assets in the portfolio.

        Returns: pd.DataFrame: A pandas dataframe representing the weights
        of various assets in the portfolio.

        Notes: The function calculates the weights of various assets
        in the portfolio by dividing the equity positions
        for each asset (as represented in the equity dataframe)
        by the total portfolio value (as represented in the nav dataframe).
        Both dataframes are assumed to have the same dimensions.
        The resulting dataframe will show the relative weight
        of each asset in the portfolio at each point in time."""
        return self.equity.apply(lambda x: x / self.nav)

    def snapshot(
        self,
        benchmark: Any = None,
        title: str = "Portfolio Summary",
        table: pd.DataFrame = None,
        aggregate: bool = False,
        log_scale: bool = False,
        label_strategy: str = "Strategy",
        label_benchmark: str = "Benchmark",
    ) -> Any:
        """
        The snapshot method creates a snapshot of the performance of a Portfolio object.
        """
        fig = make_subplots(
            rows=4,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.01,
            row_heights=[0.13, 0.47, 0.2, 0.2],
            specs=[
                [{"type": "table"}],
                [{"type": "scatter"}],
                [{"type": "scatter"}],
                [{"type": "bar"}],
            ],
        )

        # fig = make_subplots(
        #    rows=3, cols=1,
        #    shared_xaxes=True,
        #    vertical_spacing=0.03,
        #    specs=[[{"type": "table"}],
        #           [{"type": "scatter"}],
        #           [{"type": "scatter"}]]
        # )

        # display the NAV
        fig.add_trace(
            go.Scatter(x=self.nav.index, y=self.nav, name=label_strategy), row=2, col=1
        )

        # change the title of the yaxis
        fig.update_yaxes(title_text="Cumulative Return", row=2, col=1)

        # change yaxis to log scale
        if log_scale:
            fig.update_yaxes(type="log", row=2, col=1)

        # Drawdown data
        fig.add_trace(
            go.Scatter(
                x=self.drawdown.index, y=-self.drawdown, name="Drawdown", fill="tozeroy"
            ),
            row=3,
            col=1,
        )
        fig.update_yaxes(title_text="Drawdown", row=3, col=1)

        # Daily/Monthly Returns
        if aggregate:
            nav = self.nav.resample("M").last()
            returns = 100 * nav.pct_change().dropna()

            df = returns.to_frame("value").reset_index()
            df["color"] = np.where(df["value"] >= 0, "green", "red")

            fig.add_trace(
                go.Bar(
                    x=df["index"],
                    y=df["value"],
                    xperiod="M1",
                    xperiodalignment="middle",
                    marker_color=df["color"],
                    name="Monthly Returns",
                ),
                row=4,
                col=1,
            )
            fig.update_yaxes(title_text="Monthly Returns", row=4, col=1)

        else:
            returns = 100 * self.nav.pct_change().dropna()

            df = returns.to_frame("value").reset_index()
            df["color"] = np.where(df["value"] >= 0, "green", "red")

            fig.add_trace(
                go.Bar(
                    x=df["index"],
                    y=df["value"],
                    marker_color=df["color"],
                    name="Daily Returns",
                ),
                row=4,
                col=1,
            )
            fig.update_yaxes(title_text="Daily Returns", row=4, col=1)

        fig.update_traces(showlegend=True)
        fig.update_layout(height=800, title_text=title)

        if benchmark is not None:
            fig.add_trace(
                go.Scatter(x=benchmark.index, y=benchmark.values, name=label_benchmark),
                row=2,
                col=1,
            )

        if table is None:
            table = pd.DataFrame(
                index=[label_strategy],
                columns=["start", "end", "# assets", "Sharpe ratio"],
            )
            table.index.name = "Portfolio"

            table["start"][label_strategy] = self.nav.index[0].strftime("%Y-%m-%d")
            table["end"][label_strategy] = self.nav.index[-1].strftime("%Y-%m-%d")
            # table["start"][label_benchmark] = benchmark.index[0].strftime("%Y-%m-%d")
            # table["end"][label_benchmark] = benchmark.index[-1].strftime("%Y-%m-%d")
            table["# assets"][label_strategy] = len(self.assets)
            # table['# assets'][label_benchmark] =
            table["Sharpe ratio"][label_strategy] = sharpe(
                self.nav.pct_change().dropna()
            )
            # table["Sharpe ratio"][label_benchmark] = sharpe(benchmark.pct_change().dropna())

            if benchmark is not None:
                table_bench = pd.DataFrame(
                    index=[label_benchmark],
                    columns=["start", "end", "# assets", "Sharpe ratio"],
                )
                table_bench["start"][label_benchmark] = benchmark.index[0].strftime(
                    "%Y-%m-%d"
                )
                table_bench["end"][label_benchmark] = benchmark.index[-1].strftime(
                    "%Y-%m-%d"
                )
                table_bench["# assets"][label_benchmark] = ""
                table_bench["Sharpe ratio"][label_benchmark] = sharpe(
                    benchmark.pct_change().dropna()
                )

                table = pd.concat([table, table_bench], axis=0)

            table = table.reset_index()

        # print(table)
        # assert False

        # if table is not None:
        fig.add_trace(
            go.Table(
                header=dict(
                    values=list(table.columns), font=dict(size=10), align="left"
                ),
                cells=dict(
                    values=[table[column].values for column in table.columns],
                    align="left",
                ),
            ),
            row=1,
            col=1,
        )

        return fig

    @classmethod
    def from_cashpos_prices(
        cls, prices: pd.DataFrame, cashposition: pd.DataFrame, aum: float
    ):
        """Build Futures Portfolio from cashposition"""
        units = cashposition.div(prices, fill_value=0.0)
        return cls(prices=prices, units=units, aum=aum)

    @classmethod
    def from_cashpos_returns(
        cls, returns: pd.DataFrame, cashposition: pd.DataFrame, aum: float
    ):
        """Build Futures Portfolio from cashposition"""
        prices = returns2prices(returns)
        return cls.from_cashpos_prices(prices, cashposition, aum)



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from dataclasses import dataclass
from datetime import datetime

import numpy as np
import pandas as pd


@dataclass()
class State:
    """
    The state class represents the current state of a portfolio.
    It is updated within in a loop by the builder class.
    It has setter functions only for the aum, the cash, the position and the prices
    """

    _prices: pd.Series = None
    _position: pd.Series = None
    _trades: pd.Series = None
    _time: datetime = None
    _days: int = 0
    _profit: float = 0.0
    _aum: float = 0.0

    @property
    def cash(self):
        """
        The cash property returns the current amount of cash available in the portfolio.
        """
        return self.nav - self.value

    @cash.setter
    def cash(self, cash: float):
        """
        The cash property updates the amount of cash available in the portfolio.
        """
        self.aum = cash + self.value

    @property
    def nav(self) -> float:
        """
        The nav property computes the net asset value (NAV) of the portfolio,
        which is the sum of the current value of the
        portfolio as determined by the value property,
        and the current amount of cash available in the portfolio.
        """
        # assert np.isclose(self.value + self.cash, self.aum), f"{self.value + self.cash} != {self.aum}"
        # return self.value + self.cash
        return self.aum

    @property
    def value(self) -> float:
        """
        The value property computes the value of the portfolio at the current
        time taking into account the current holdings and current prices.
        If the value cannot be computed due to missing positions
        (they might be still None), zero is returned instead.
        """
        return self.cashposition.sum()

    @property
    def cashposition(self):
        """
        The `cashposition` property computes the cash position of the portfolio,
        which is the amount of cash in the portfolio as a fraction of the total portfolio value.
        """
        return self.prices * self.position

    @property
    def position(self):
        """Get the current position. If the position is not yet set, then return an empty series."""
        if self._position is None:
            return pd.Series(index=self.assets, dtype=float)

        return self._position

    @position.setter
    def position(self, position: np.array):
        """
        Update the position of the state. Computes the required trades
        and updates other quantities (e.g. cash) accordingly.
        """
        # update the position
        position = pd.Series(index=self.assets, data=position)

        # compute the trades (can be fractional)
        self._trades = position.subtract(self.position, fill_value=0.0)

        # update only now as otherwise the trades would be wrong
        self._position = position

    @property
    def gmv(self):
        """
        gross market value, e.g. abs(short) + long
        """
        return self.cashposition.abs().sum()

    @property
    def time(self):
        """
        The time property returns the current time of the state.
        """
        return self._time

    @time.setter
    def time(self, time: datetime):
        """
        Update the time of the state. Computes the number of days between
        the new and the previous time.
        """
        if self.time is None:
            self._days = 0
            self._time = time
        else:
            self._days = (time - self.time).days
            self._time = time

    @property
    def days(self):
        """Number of days between the current and the previous time. Most useful
        for computing the interest when holding cash."""
        return self._days

    @property
    def assets(self) -> pd.Index:
        """
        The assets property returns the assets currently in the portfolio.
        """
        if self._prices is None:
            return pd.Index(data=[], dtype=str)

        return self.prices.dropna().index

    @property
    def trades(self):
        """
        The trades property returns the trades currently needed to reach the position.
        Most helpful when computing the trading costs following the move to
        a new position.
        """
        return self._trades

    @property
    def mask(self):
        """construct true/false mask for assets with missing prices"""
        if self._prices is None:
            return np.array([])

        return np.isfinite(self.prices.values)

    @property
    def prices(self):
        """Get the current prices"""
        if self._prices is None:
            return pd.Series(dtype=float)
        return self._prices

    @prices.setter
    def prices(self, prices):
        """
        The prices property updates the prices of the assets in the portfolio.
        Also updates the aum of the portfolio by the profit achieved.
        """
        value_before = (self.prices * self.position).sum()  # self.cashposition.sum()
        value_after = (prices * self.position).sum()

        self._prices = prices
        self._profit = value_after - value_before
        self.aum += self.profit

    @property
    def profit(self):
        """
        Following the price update the state updates the profit achieved
        between the previous and the current prices.
        """
        return self._profit

    @property
    def aum(self):
        """
        The aum property returns the current assets under management (AUM) of the portfolio.
        """
        return self._aum

    @aum.setter
    def aum(self, aum):
        """
        Setter for the aum property. It updates the aum property.
        """
        self._aum = aum

    @property
    def weights(self) -> pd.Series:
        """
        The weights property computes the weighting of each asset in the current
        portfolio as a fraction of the total portfolio value (nav).

        Returns:

        a pandas series object containing the weighting of each asset as a
        fraction of the total portfolio value. If the positions are still
        missing, then a series of zeroes is returned.
        """
        assert np.isclose(self.nav, self.aum), f"{self.nav} != {self.aum}"
        return self.cashposition / self.nav

    @property
    def leverage(self) -> float:
        """
        The `leverage` property computes the leverage of the portfolio,
        which is the sum of the absolute values of the portfolio weights.
        """
        return float(self.weights.abs().sum())
